@@ -36,14 +36,14 @@ logger = logging.getLogger(__name__)
 
 DEFAULT_ATTEMPTS = 3
 
-INITIAL_ATTRS = (
+INITIAL_ATTRS = {
     "meta",
     "udsk",
     "dsk",
     "date_time_zone",
     "firmware",
-)
-UPDATE_ATTRS = (
+}
+UPDATE_ATTRS = {
     "name",
     "led_colour",
     "current_temp",
@@ -53,7 +53,8 @@ UPDATE_ATTRS = (
     "liquid_level",
     "liquid_state",
     "battery_voltage",
-)
+}
+EXTRA_ATTRS = {'dsk', 'udsk', 'battery_voltage', 'date_time_zone'}
 
 
 class EmberMugConnection:
@@ -68,7 +69,10 @@ class EmberMugConnection:
 
         self._queued_updates: set[str] = set()
         self._latest_events: dict[int, float] = {}
+        self._initial_attrs = INITIAL_ATTRS if mug.include_extra else (INITIAL_ATTRS - EXTRA_ATTRS)
+        self._update_attrs = UPDATE_ATTRS if mug.include_extra else (UPDATE_ATTRS - EXTRA_ATTRS)
 
+        logger.debug("New mug connection initialized.")
         self._client_kwargs = {**kwargs}
         if adapter:
             if USES_BLUEZ is False:
@@ -85,6 +89,7 @@ class EmberMugConnection:
 
     def set_device(self, ble_device: BLEDevice) -> None:
         """Set the ble device."""
+        logger.debug("Set new device from %s to %s", self._device, ble_device)
         self._device = ble_device
 
     async def ensure_connection(self) -> None:
@@ -97,6 +102,7 @@ class EmberMugConnection:
             if self._client is not None and self._client.is_connected is True:
                 return
             try:
+                logger.debug("Establishing a new connection")
                 self._client = await establish_connection(
                     client_class=BleakClient,
                     device=self._device,
@@ -117,7 +123,7 @@ class EmberMugConnection:
                 # workaround for Home Assistant ESPHome Proxy backend which does not allow pairing.
                 logger.warning(
                     'Pairing not implemented. '
-                    'If your mug is still in pairing mode (blinking blue) tap the button on the bottom to exit.'
+                    'If your mug is still in pairing mode (blinking blue) tap the button on the bottom to exit.',
                 )
             await self.update_initial()
             await self.subscribe()
@@ -132,13 +138,14 @@ class EmberMugConnection:
         """Helper to write characteristic to Mug."""
         try:
             await self._client.write_gatt_char(characteristic.uuid, data)
-            logger.debug("Wrote %s to attribute '%s'", characteristic, data)
+            logger.debug("Wrote '%s' to attribute '%s'", data, characteristic)
         except BleakError as e:
             logger.error("Failed to write '%s' to attribute '%s': %s", data, characteristic, e)
             raise
 
     async def disconnect(self) -> None:
         """Disconnect from mug and stop listening to notifications."""
+        logger.debug("Disconnect called")
         if self._client and self._client.is_connected is True:
             async with self._connect_lock:
                 await self.unsubscribe()
@@ -146,6 +153,7 @@ class EmberMugConnection:
 
     def _disconnect_callback(self, client: BleakClient) -> None:
         """Disconnect from device."""
+        logger.debug("Disconnect callback called")
         asyncio.create_task(self.disconnect())
 
     def _fire_callbacks(self) -> None:
@@ -171,7 +179,10 @@ class EmberMugConnection:
         return self
 
     async def __aexit__(
-        self, exc_type: type[BaseException], exc_val: BaseException, exc_tb: TracebackType | None
+        self,
+        exc_type: type[BaseException],
+        exc_val: BaseException,
+        exc_tb: TracebackType | None,
     ) -> None:
         """Cleanup on exit."""
         await self.disconnect()
@@ -235,7 +246,11 @@ class EmberMugConnection:
 
     async def get_udsk(self) -> str:
         """Get mug udsk from gatt."""
-        return decode_byte_string(await self._read(MugCharacteristic.UDSK))
+        try:
+            return decode_byte_string(await self._read(MugCharacteristic.UDSK))
+        except BleakError as e:
+            logger.debug('Unable to read UDSK: %s', e)
+        return ''
 
     async def set_udsk(self, udsk: str) -> None:
         """Attempt to write udsk."""
@@ -244,7 +259,11 @@ class EmberMugConnection:
 
     async def get_dsk(self) -> str:
         """Get mug dsk from gatt."""
-        return decode_byte_string(await self._read(MugCharacteristic.DSK))
+        try:
+            return decode_byte_string(await self._read(MugCharacteristic.DSK))
+        except BleakError as e:
+            logger.debug('Unable to read DSK: %s', e)
+        return ''
 
     async def get_temperature_unit(self) -> TemperatureUnit:
         """Get mug temp unit."""
@@ -284,13 +303,13 @@ class EmberMugConnection:
 
     async def update_initial(self) -> list[Change]:
         """Update attributes that don't normally change and don't need to be regularly updated."""
-        return await self._update_multiple(INITIAL_ATTRS)
+        return await self._update_multiple(self._initial_attrs)
 
     async def update_all(self) -> list[Change]:
         """Update all standard attributes."""
-        return await self._update_multiple(UPDATE_ATTRS)
+        return await self._update_multiple(self._update_attrs)
 
-    async def _update_multiple(self, attrs: tuple[str, ...]) -> list[Change]:
+    async def _update_multiple(self, attrs: set[str]) -> list[Change]:
         """Helper to update a list of attributes from the mug."""
         logger.debug('Updating the following attributes: %s', attrs)
         changes = self.mug.update_info(**{attr: await getattr(self, f"get_{attr}")() for attr in attrs})
@@ -311,7 +330,7 @@ class EmberMugConnection:
             self._fire_callbacks()
         return changes
 
-    def _notify_callback(self, characteristic: int | BleakGATTCharacteristic, data: bytearray) -> None:
+    def _notify_callback(self, characteristic: BleakGATTCharacteristic, data: bytearray) -> None:
         """Push events from the mug to indicate changes."""
         event_id = data[0]
         now = time()
@@ -350,6 +369,7 @@ class EmberMugConnection:
 
     async def unsubscribe(self) -> None:
         """Unsubscribe from Mug notifications."""
+        logger.debug("Unsubscribe called")
         with contextlib.suppress(BleakError):
             await self._client.stop_notify(MugCharacteristic.PUSH_EVENT.uuid)
 
