@@ -1,11 +1,13 @@
 """Tests for `ember_mug.connection`."""
+from __future__ import annotations
+
 from unittest.mock import AsyncMock, Mock, patch
 
 import pytest
 from bleak import BleakError
 from bleak.backends.device import BLEDevice
 
-from ember_mug.connection import INITIAL_ATTRS, UPDATE_ATTRS, EmberMugConnection
+from ember_mug.connection import EXTRA_ATTRS, INITIAL_ATTRS, UPDATE_ATTRS, EmberMugConnection
 from ember_mug.consts import MugCharacteristic, TemperatureUnit
 from ember_mug.data import Colour
 
@@ -24,7 +26,6 @@ def test_adapter_without_bluez(ember_mug):
 
 @patch('ember_mug.connection.EmberMugConnection.update_initial')
 @patch('ember_mug.connection.establish_connection')
-@pytest.mark.asyncio
 async def test_connect(mug_update_initial, mock_establish_connection, mug_connection):
     # Already connected
     mug_connection._client = AsyncMock()
@@ -46,9 +47,10 @@ async def test_connect(mug_update_initial, mock_establish_connection, mug_connec
 
 @patch('ember_mug.connection.logger')
 @patch('ember_mug.connection.establish_connection')
-@pytest.mark.asyncio
 async def test_connect_error(
-    mock_establish_connection: Mock, mock_logger: Mock, mug_connection: EmberMugConnection
+    mock_establish_connection: AsyncMock,
+    mock_logger: Mock,
+    mug_connection: EmberMugConnection,
 ) -> None:
     mug_connection._client = None  # type: ignore[assignment]
     mock_establish_connection.side_effect = BleakError
@@ -62,24 +64,43 @@ async def test_connect_error(
 
 @patch('ember_mug.connection.logger')
 @patch('ember_mug.connection.establish_connection')
-@pytest.mark.asyncio
-async def test_pairing_exceptions(
-    mock_establish_connection: Mock, mock_logger: Mock, mug_connection: EmberMugConnection
+async def test_pairing_exceptions_esphome(
+    mock_establish_connection: AsyncMock,
+    mock_logger: Mock,
+    mug_connection: EmberMugConnection,
 ) -> None:
     mock_client = AsyncMock()
     mock_client.pair.side_effect = NotImplementedError
     mock_establish_connection.return_value = mock_client
-    mug_connection.update_initial = AsyncMock()  # type: ignore[assignment]
-    mug_connection.subscribe = AsyncMock()  # type: ignore[assignment]
-    await mug_connection.ensure_connection()
+    with patch.multiple(
+        mug_connection,
+        update_initial=AsyncMock(),
+        subscribe=AsyncMock(),
+    ):
+        await mug_connection.ensure_connection()
     mock_logger.warning.assert_called_with(
         'Pairing not implemented. '
-        'If your mug is still in pairing mode (blinking blue) tap the button on the bottom to exit.'
+        'If your mug is still in pairing mode (blinking blue) tap the button on the bottom to exit.',
     )
 
 
-@pytest.mark.asyncio
-async def test_disconnect(mug_connection):
+@patch('ember_mug.connection.establish_connection')
+async def test_pairing_exceptions(
+    mock_establish_connection: AsyncMock,
+    mug_connection: EmberMugConnection,
+) -> None:
+    mock_client = AsyncMock()
+    mock_client.pair.side_effect = BleakError
+    mock_establish_connection.return_value = mock_client
+    with patch.multiple(
+        mug_connection,
+        update_initial=AsyncMock(),
+        subscribe=AsyncMock(),
+    ):
+        await mug_connection.ensure_connection()
+
+
+async def test_disconnect(mug_connection: EmberMugConnection) -> None:
     mug_connection._client = AsyncMock()
 
     mug_connection._client.is_connected = False
@@ -91,6 +112,75 @@ async def test_disconnect(mug_connection):
     mug_connection._client.disconnect.assert_called()
 
 
+@patch('ember_mug.connection.logger')
+@patch('ember_mug.connection.asyncio')
+def test_disconnect_callback(
+    mock_asyncio: AsyncMock,
+    mock_logger: Mock,
+    mug_connection: EmberMugConnection,
+) -> None:
+    mug_connection._disconnect_callback(AsyncMock())
+    mock_logger.debug.assert_called_with("Disconnect callback called")
+    mock_asyncio.create_task.assert_called_once()
+
+
+@patch('ember_mug.connection.logger')
+async def test_read(
+    mock_logger: Mock,
+    mug_connection: EmberMugConnection,
+) -> None:
+    mug_connection._client = AsyncMock()
+    mug_connection._client.read_gatt_char.return_value = b'TEST'
+    await mug_connection._read(MugCharacteristic.MUG_NAME)
+    mug_connection._client.read_gatt_char.assert_called_with(
+        MugCharacteristic.MUG_NAME.uuid,
+    )
+    mock_logger.debug.assert_called_with(
+        "Read attribute '%s' with value '%s'",
+        MugCharacteristic.MUG_NAME,
+        b'TEST',
+    )
+
+
+@patch('ember_mug.connection.logger')
+async def test_write(
+    mock_logger: Mock,
+    mug_connection: EmberMugConnection,
+) -> None:
+    mug_connection._client = AsyncMock()
+    test_name = bytearray(b'TEST')
+    await mug_connection._write(
+        MugCharacteristic.MUG_NAME,
+        test_name,
+    )
+    mug_connection._client.write_gatt_char.assert_called_with(
+        MugCharacteristic.MUG_NAME.uuid,
+        test_name,
+    )
+    mock_logger.debug.assert_called_with(
+        "Wrote '%s' to attribute '%s'",
+        test_name,
+        MugCharacteristic.MUG_NAME,
+    )
+
+    mug_connection._client = AsyncMock()
+    mug_connection._client.write_gatt_char.side_effect = BleakError
+    with pytest.raises(BleakError):
+        await mug_connection._write(
+            MugCharacteristic.MUG_NAME,
+            test_name,
+        )
+    mug_connection._client.write_gatt_char.assert_called_with(
+        MugCharacteristic.MUG_NAME.uuid,
+        test_name,
+    )
+    msg, data, char, exception = mock_logger.error.mock_calls[0].args
+    assert msg == "Failed to write '%s' to attribute '%s': %s"
+    assert data == test_name
+    assert char == MugCharacteristic.MUG_NAME
+    assert isinstance(exception, BleakError)
+
+
 def test_set_device(mug_connection: EmberMugConnection) -> None:
     new_device = BLEDevice(
         address='BA:36:a5:be:88:cb',
@@ -101,7 +191,6 @@ def test_set_device(mug_connection: EmberMugConnection) -> None:
     assert mug_connection.mug.device.address == new_device.address
 
 
-@pytest.mark.asyncio
 async def test_get_mug_meta(mug_connection):
     mug_connection._client.read_gatt_char.return_value = b'Yw====-ABCDEFGHIJ'
     meta = await mug_connection.get_meta()
@@ -110,7 +199,6 @@ async def test_get_mug_meta(mug_connection):
     mug_connection._client.read_gatt_char.assert_called_once_with(MugCharacteristic.MUG_ID.uuid)
 
 
-@pytest.mark.asyncio
 async def test_get_mug_battery(mug_connection):
     mug_connection._client.read_gatt_char.return_value = b'5\x01'
     battery = await mug_connection.get_battery()
@@ -119,7 +207,6 @@ async def test_get_mug_battery(mug_connection):
     mug_connection._client.read_gatt_char.assert_called_once_with(MugCharacteristic.BATTERY.uuid)
 
 
-@pytest.mark.asyncio
 async def test_get_mug_led_colour(mug_connection):
     mug_connection._client.read_gatt_char.return_value = b'\xf4\x00\xa1\xff'
     colour = await mug_connection.get_led_colour()
@@ -127,62 +214,56 @@ async def test_get_mug_led_colour(mug_connection):
     mug_connection._client.read_gatt_char.assert_called_once_with(MugCharacteristic.LED.uuid)
 
 
-@pytest.mark.asyncio
 async def test_set_mug_led_colour(mug_connection):
     mug_connection.ensure_connection = AsyncMock()
     await mug_connection.set_led_colour(Colour(244, 0, 161))
     mug_connection.ensure_connection.assert_called_once()
     mug_connection._client.write_gatt_char.assert_called_once_with(
-        MugCharacteristic.LED.uuid, bytearray(b'\xf4\x00\xa1\xff')
+        MugCharacteristic.LED.uuid,
+        bytearray(b'\xf4\x00\xa1\xff'),
     )
 
 
-@pytest.mark.asyncio
 async def test_get_mug_target_temp(mug_connection):
     mug_connection._client.read_gatt_char.return_value = b'\xcd\x15'
     assert (await mug_connection.get_target_temp()) == 55.81
     mug_connection._client.read_gatt_char.assert_called_once_with(MugCharacteristic.TARGET_TEMPERATURE.uuid)
 
 
-@pytest.mark.asyncio
 async def test_set_mug_target_temp(mug_connection):
     mug_connection.ensure_connection = AsyncMock()
     await mug_connection.set_target_temp(55.81)
     mug_connection.ensure_connection.assert_called_once()
     mug_connection._client.write_gatt_char.assert_called_once_with(
-        MugCharacteristic.TARGET_TEMPERATURE.uuid, bytearray(b'\xcd\x15')
+        MugCharacteristic.TARGET_TEMPERATURE.uuid,
+        bytearray(b'\xcd\x15'),
     )
 
 
-@pytest.mark.asyncio
 async def test_get_mug_current_temp(mug_connection):
     mug_connection._client.read_gatt_char.return_value = b'\xcd\x15'
     assert (await mug_connection.get_current_temp()) == 55.81
     mug_connection._client.read_gatt_char.assert_called_once_with(MugCharacteristic.CURRENT_TEMPERATURE.uuid)
 
 
-@pytest.mark.asyncio
 async def test_get_mug_liquid_level(mug_connection):
     mug_connection._client.read_gatt_char.return_value = b'\n'
     assert (await mug_connection.get_liquid_level()) == 10
     mug_connection._client.read_gatt_char.assert_called_once_with(MugCharacteristic.LIQUID_LEVEL.uuid)
 
 
-@pytest.mark.asyncio
 async def test_get_mug_liquid_state(mug_connection):
     mug_connection._client.read_gatt_char.return_value = b'\x06'
     assert (await mug_connection.get_liquid_state()) == 6
     mug_connection._client.read_gatt_char.assert_called_once_with(MugCharacteristic.LIQUID_STATE.uuid)
 
 
-@pytest.mark.asyncio
 async def test_get_mug_name(mug_connection):
     mug_connection._client.read_gatt_char.return_value = b'Mug Name'
     assert (await mug_connection.get_name()) == 'Mug Name'
     mug_connection._client.read_gatt_char.assert_called_once_with(MugCharacteristic.MUG_NAME.uuid)
 
 
-@pytest.mark.asyncio
 async def test_set_mug_name(mug_connection):
     mug_connection.ensure_connection = AsyncMock()
     with pytest.raises(ValueError):
@@ -191,28 +272,27 @@ async def test_set_mug_name(mug_connection):
     await mug_connection.set_name('Mug name')
     mug_connection.ensure_connection.assert_called()
     mug_connection._client.write_gatt_char.assert_called_once_with(
-        MugCharacteristic.MUG_NAME.uuid, bytearray(b'Mug name')
+        MugCharacteristic.MUG_NAME.uuid,
+        bytearray(b'Mug name'),
     )
 
 
-@pytest.mark.asyncio
 async def test_get_mug_udsk(mug_connection):
     mug_connection._client.read_gatt_char.return_value = b'abcd12345'
     assert (await mug_connection.get_udsk()) == 'YWJjZDEyMzQ1'
     mug_connection._client.read_gatt_char.assert_called_once_with(MugCharacteristic.UDSK.uuid)
 
 
-@pytest.mark.asyncio
 async def test_set_mug_udsk(mug_connection):
     mug_connection.ensure_connection = AsyncMock()
     await mug_connection.set_udsk('abcd12345')
     mug_connection.ensure_connection.assert_called_once()
     mug_connection._client.write_gatt_char.assert_called_once_with(
-        MugCharacteristic.UDSK.uuid, bytearray(b'YWJjZDEyMzQ1')
+        MugCharacteristic.UDSK.uuid,
+        bytearray(b'YWJjZDEyMzQ1'),
     )
 
 
-@pytest.mark.asyncio
 async def test_get_mug_dsk(mug_connection):
     mug_connection._client.read_gatt_char.return_value = b'abcd12345'
     assert (await mug_connection.get_dsk()) == 'YWJjZDEyMzQ1'
@@ -221,7 +301,6 @@ async def test_get_mug_dsk(mug_connection):
     mug_connection._client.read_gatt_char.assert_called_with(MugCharacteristic.DSK.uuid)
 
 
-@pytest.mark.asyncio
 async def test_get_mug_temperature_unit(mug_connection):
     mug_connection._client.read_gatt_char.return_value = b'\x01'
     assert (await mug_connection.get_temperature_unit()) == TemperatureUnit.FAHRENHEIT
@@ -232,17 +311,16 @@ async def test_get_mug_temperature_unit(mug_connection):
     mug_connection._client.read_gatt_char.assert_called_once_with(MugCharacteristic.TEMPERATURE_UNIT.uuid)
 
 
-@pytest.mark.asyncio
 async def test_set_mug_temperature_unit(mug_connection):
     mug_connection.ensure_connection = AsyncMock()
     await mug_connection.set_temperature_unit(TemperatureUnit.CELSIUS)
     mug_connection.ensure_connection.assert_called_once()
     mug_connection._client.write_gatt_char.assert_called_once_with(
-        MugCharacteristic.TEMPERATURE_UNIT.uuid, bytearray(b'\x00')
+        MugCharacteristic.TEMPERATURE_UNIT.uuid,
+        bytearray(b'\x00'),
     )
 
 
-@pytest.mark.asyncio
 async def test_mug_ensure_correct_unit(mug_connection):
     mug_connection.mug.temperature_unit = TemperatureUnit.CELSIUS
     mug_connection.mug.use_metric = True
@@ -254,14 +332,12 @@ async def test_mug_ensure_correct_unit(mug_connection):
     mug_connection.set_temperature_unit.assert_called_with(TemperatureUnit.CELSIUS)
 
 
-@pytest.mark.asyncio
 async def test_get_mug_battery_voltage(mug_connection):
     mug_connection._client.read_gatt_char.return_value = b'\x01'
     assert (await mug_connection.get_battery_voltage()) == 1
     mug_connection._client.read_gatt_char.assert_called_once_with(MugCharacteristic.CONTROL_REGISTER_DATA.uuid)
 
 
-@pytest.mark.asyncio
 async def test_get_mug_date_time_zone(mug_connection):
     mug_connection._client.read_gatt_char.return_value = b'c\x0f\xf6\x00'
     date_time = await mug_connection.get_date_time_zone()
@@ -269,7 +345,6 @@ async def test_get_mug_date_time_zone(mug_connection):
     mug_connection._client.read_gatt_char.assert_called_once_with(MugCharacteristic.DATE_TIME_AND_ZONE.uuid)
 
 
-@pytest.mark.asyncio
 async def test_read_firmware(mug_connection):
     mug_connection._client.read_gatt_char.return_value = b'c\x01\x80\x00\x12\x00'
     firmware = await mug_connection.get_firmware()
@@ -279,23 +354,34 @@ async def test_read_firmware(mug_connection):
     mug_connection._client.read_gatt_char.assert_called_once_with(MugCharacteristic.FIRMWARE.uuid)
 
 
-@pytest.mark.asyncio
 async def test_mug_update_initial(mug_connection):
+    no_extra = INITIAL_ATTRS - EXTRA_ATTRS
+
     mug_connection._update_multiple = AsyncMock(return_value={})
     mug_connection.ensure_connection = AsyncMock()
     assert (await mug_connection.update_initial()) == {}
+    mug_connection._update_multiple.assert_called_once_with(no_extra)
+
+    # Try with extra
+    mug_connection._update_multiple.reset_mock()
+    mug_connection._initial_attrs = INITIAL_ATTRS
+    assert (await mug_connection.update_initial()) == {}
     mug_connection._update_multiple.assert_called_once_with(INITIAL_ATTRS)
 
 
-@pytest.mark.asyncio
 async def test_mug_update_all(mug_connection):
     mug_connection._update_multiple = AsyncMock(return_value={})
     mug_connection.ensure_connection = AsyncMock()
     assert (await mug_connection.update_all()) == {}
+    mug_connection._update_multiple.assert_called_once_with(UPDATE_ATTRS - EXTRA_ATTRS)
+
+    # Try with extras
+    mug_connection._update_multiple.reset_mock()
+    mug_connection._update_attrs = UPDATE_ATTRS
+    assert (await mug_connection.update_all()) == {}
     mug_connection._update_multiple.assert_called_once_with(UPDATE_ATTRS)
 
 
-@pytest.mark.asyncio
 async def test_mug_update_multiple(mug_connection):
     mug_connection.get_name = AsyncMock(return_value='name')
     mug_connection.mug.update_info = AsyncMock()
@@ -303,7 +389,6 @@ async def test_mug_update_multiple(mug_connection):
     mug_connection.mug.update_info.assert_called_once_with(name='name')
 
 
-@pytest.mark.asyncio
 async def test_mug_update_queued_attributes(mug_connection):
     mug_connection._queued_updates = set()
     assert (await mug_connection.update_queued_attributes()) == []
@@ -314,23 +399,25 @@ async def test_mug_update_queued_attributes(mug_connection):
     mug_connection.mug.update_info.assert_called_once_with(name='name')
 
 
-def test_mug_notify_callback(mug_connection):
-    mug_connection._notify_callback(1, b'\x02')
-    mug_connection._notify_callback(1, b'\x02')
+def test_mug_notify_callback(mug_connection: EmberMugConnection) -> None:
+    gatt_char = AsyncMock()
+    mug_connection._notify_callback(gatt_char, bytearray(b'\x01'))
+    mug_connection._notify_callback(gatt_char, bytearray(b'\x02'))
     assert 2 in mug_connection._latest_events
-    mug_connection._notify_callback(1, b'\x04')
+    mug_connection._notify_callback(gatt_char, bytearray(b'\x04'))
     assert 4 in mug_connection._latest_events
-    mug_connection._notify_callback(1, b'\x05')
+    mug_connection._notify_callback(gatt_char, bytearray(b'\x05'))
     assert 5 in mug_connection._latest_events
-    mug_connection._notify_callback(1, b'\x06')
+    mug_connection._notify_callback(gatt_char, bytearray(b'\x06'))
     assert 6 in mug_connection._latest_events
-    mug_connection._notify_callback(1, b'\x07')
+    mug_connection._notify_callback(gatt_char, bytearray(b'\x07'))
     assert 7 in mug_connection._latest_events
-    mug_connection._notify_callback(1, b'\x08')
+    mug_connection._notify_callback(gatt_char, bytearray(b'\x08'))
     assert 8 in mug_connection._latest_events
     callback = Mock()
-    mug_connection.register_callback(callback)
-    mug_connection._notify_callback(1, b'\x09')
+    unregister = mug_connection.register_callback(callback)
+    assert callback in mug_connection._callbacks
+    mug_connection._notify_callback(gatt_char, bytearray(b'\x09'))
     assert 9 in mug_connection._latest_events
     callback.assert_not_called()
     assert mug_connection._queued_updates == {
@@ -342,8 +429,11 @@ def test_mug_notify_callback(mug_connection):
         "battery_voltage",
     }
     mug_connection._latest_events = {}
-    mug_connection._notify_callback(1, b'\x02')
+    mug_connection._notify_callback(gatt_char, bytearray(b'\x02'))
     callback.assert_called_once()
     callback.reset_mock()
-    mug_connection._notify_callback(1, b'\x02')
+    mug_connection._notify_callback(gatt_char, bytearray(b'\x02'))
     callback.assert_not_called()
+    # Remove callback
+    unregister()
+    assert callback not in mug_connection._callbacks
